// ***************************************************************************************
// File: application.h
// Description: declarations of the Nirva application classes
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: An application is a set of defined rules and parameters
//           A session must be opened in an application context.
// ***************************************************************************************

#ifndef _APPLICATION_H
#define _APPLICATION_H

// Classes declared in this file
class NvRedirection;
class NvRedirectionList;
class NvApplication;
class NvApplicationList;


#include "task.h"
#include "transaction.h"
#include "scheduler.h"
#include "container.h"
#include "locks.h"
#include "security.h"
#include "listener.h"
#include "nidget.h"
#include "test.h"
#include <string>


#define NV_APPLICATION_DEFAULT_NAME				"NVDEF"

// Authentication modes
#define NV_AUTHENTICATION_NIRVA			0
#define NV_AUTHENTICATION_SSO			1
#define NV_AUTHENTICATION_NIRVA_SSO		2
#define NV_AUTHENTICATION_MAXVALUE		2


// Command filter use modes
#define NV_COMMAND_FILTER_NEVER			0
#define NV_COMMAND_FILTER_ALWAYS		1
#define NV_COMMAND_FILTER_BYSECU		2
#define NV_COMMAND_FILTER_MAXVALUE		2


/////////////////////////////////////////////////////////////////////////////
// NvRedirection class
class NvRedirection : public NvObject
{
public:
		
	std::string m_sAddress;						// Address
	bool m_bEnable;								// Enable

	NvRedirection();
	virtual ~NvRedirection();
};


/////////////////////////////////////////////////////////////////////////////
// NvRedirectionList class
class NvRedirectionList : public NvObject
{
private:
	NvRedirection* InternalGetRedirection(const char *Address);

public:
	std::map<std::string, NvRedirection*> m_mRedirections;	// Redirection list
	NvApplication *Application;								// Application
	
	NvRedirectionList();
	virtual ~NvRedirectionList();

	// Removes all redirections from the list
	void Clear();

	// Load the redirection list from registry
	bool Load();

	// Load the redirection list from registry
	bool Save();

	// Get list for http
	bool GetRedirections(std::vector<std::string> &List);

	// Get redirection info
	bool GetRedirectionInfo(const char *Address, NvObjTable *RedirectionInfo);

	// Add a redirection
	bool AddRedirection(const char *Address, bool bLockIt = true);

	// Remove a redirection
	bool RemoveRedirection(const char *Address);

	// Enable a redirection
	bool EnableRedirection(const char *Address, bool Enable, bool bLockIt = true);

};

/////////////////////////////////////////////////////////////////////////////
// NvApplication class
class NvApplication : public NvObject
{
private:
	int InUse;									// In use counter
	NvMutex ProtectInUse;						// protect locking of application for updating the in use counter
	bool Initialized;
	bool CreateApplicationDirectories();		// Creates application directories if they don't exist

public:
	std::string m_sName;						// Application name
	std::string m_sDesc;						// Application description
	int m_iDefaultTimeOut;						// Application default time out
	int m_iAuthMode;							// Application authentication mode
	bool m_bNirvaIsSsoUser;						// Set to true when Nirva user is sso user
	NvRegistryContainer AppRegistry;			// Application parameters (registry)
	bool m_bEnable;								// Application enabled or not
	bool m_bStarted;							// Application started or not
	bool m_bInStartStop;						// Application is in start or stop step (used by scheduler)
	std::string m_sWorkDir;						// Temporary file directory (work dir)
	std::string m_sRegistryDir;					// Registry directory
	std::string m_sDocDir;						// Doc directory
	std::string m_sFileDir;						// File directory
	std::string m_sLogDir;						// Log directory
	std::string m_sWrootDir;					// Web directory
	std::string m_sProcDir;						// Procedure directory
	std::string m_sStartPage;					// Html starting page
	NvLockList LockList;						// application lock list
	NvSemaList SemaList;						// application semaphore list
	NvTransactionList Transactions;				// Transaction list
	NvSchedulerTaskList Tasks;					// Task list
	NvObjIndStringList Sessions;				// List of sessions actually connected to the application
	NvDescriptionContainer m_Info;				// Application information
	NvLogList Logs;								// Application logs
	bool bInInstall;							// In installation process
	NvSecurity Security;						// Security object that maintains users and roles
	std::string m_sSecurityApplication;			// Name of the application to get security from ("(SYSTEM)" for system security)
	bool bUseScheduledTransac;					// Use scheduled transactions or not
	bool bAllowWebXml;							// Allow display of xml in browser
	bool bAllowNotOwnedSession;					// Allow clients to send commands to a session they don't own
	NvContainerList	Containers;					// Application container list
	NvListenerList Listeners;					// Application listeners
	NvRegistryContainerList Registries;			// Application user registries
	NvTestSetList TestSets;						// Test sets
	NvRedirectionList Redirections;				// Redirection server adresses
	NvNidgetFramework NidgetFramework;			// Nidget framework
	bool m_bUseFramework;						// Use Nidget framework
	NvCommandFilter CommandClientFilter;		// Command filter for client command
	NvCommandFilter CommandWebFilter;			// Command filter for web command
	int m_iCommandClientFilterMode;				// Client command filter mode
	int m_iCommandWebFilterMode;				// Web command filter mode
	std::string m_sRequiredServices;			// List of services required for this application
	
		
	NvApplication();
	virtual ~NvApplication();

	bool Init();
	void Cleanup();

	void IncInuse();
	void DecInUse();
	int GetInUse();

	bool Enable();
	bool IsEnable();
	void SetInStartStop(bool Value);
	bool IsInStartStop();
	bool IsFirstSession();
	bool Start(bool EnableIt = false, bool WithListeners = true);
	bool Stop(const char *SessionId = NULL, bool DisableIt = false);

	bool SetParameters(const char *Name, const char *Desc, int DefaultTimeOut = 300, const char *SecuApp = NULL, bool ScheduledTransac = true, bool AllowWebXml = true, bool AllowNotOwnedSession = true, int AuthMode = NV_AUTHENTICATION_NIRVA, bool NirvaIsSsoUser = false, const char* ClientFilterString = NULL, const char* WebFilterString = NULL, int CommandClientFilterMode = NV_COMMAND_FILTER_NEVER, int CommandWebFilterMode = NV_COMMAND_FILTER_NEVER);
	bool Package(const char *DescriptionFile, const char *PackageFile, std::string& Info);
	bool Install(const char *PackageFile, std::string& Info);
	bool CreateLicenseFile();					// Creates the license file if not existing

	std::string GetName();
	std::string GetDesc();
	int GetDefaultTimeOut();
	int GetAuthMode();
	bool GetNirvaIsSsoUser();
	std::string GetSecuApp();
	bool GetUseScheduledTransac();
	bool GetAllowWebXml();
	bool GetAllowNotOwnedSession();
	int GetCommandClientFilterMode();
	int GetCommandWebFilterMode();
	std::string GetCommandClientFilter();
	std::string GetCommandWebFilter();
	std::string GetWorkDir();
	std::string GetRegistryDir();
	std::string GetDocDir();
	std::string GetFileDir();
	std::string GetLogDir();
	std::string GetWrootDir();
	std::string GetProcDir();
	std::string GetAppNameForFiles();
	std::string GetStartPage();
	std::string GetRequiredServices();
	bool GetUseFramework();
	bool SetApplicationFile(NvObjFile *File);
	bool LoadInfo();
	bool GetInfoEx(NvObjIndStringList *ApplicationInfo);
	static void CleanAppName(std::string& Name);
	int Login(const char *SessionId, const char *UserName, const char *UserPassword, const char *NewPassword, const char *NewPasswordConfirm, bool bWithNewPassword, bool bWithNewPasswordConfirm, NvObjIndStringList *UserPermissions, std::string& UserFullName, bool ControlPassword = true);
	int ExternalLogin(const char *SessionId, const char *UserName, const char *UserPassword, const char *NewPassword, const char *NewPasswordConfirm, bool bWithNewPassword, bool bWithNewPasswordConfirm, NvObjIndStringList *UserPermissions, std::string& UserFullName, bool ControlPassword = true);
	int ServiceLogin(const char *SessionId, const char *UserName, const char *UserPassword, const char *NewPassword, const char *NewPasswordConfirm, bool bWithNewPassword, bool bWithNewPasswordConfirm, NvObjIndStringList *UserPermissions, std::string& UserFullName, bool ControlPassword = true);
	bool Logout(const char *SessionId);
	NvSecurity* GetSecurity(std::string& AppName);
	bool IsLocalSecurity();
	std::string GetServiceSecurity();

	// Get sessions ids connected to the application
	bool GetSessions(NvObjStringList *ConnectedSessions);

	// Get application permissions
	bool GetPermissions(NvObjTable *ApplicationPermissions);

	// CheckRequiredService
	bool IsRequiredService(std::string& ServiceName);
};


/////////////////////////////////////////////////////////////////////////////
// NvApplicationList class
class NvApplicationList : public NvObject
{
private:
	NvApplication* InternalGetApplication(const char *Name, int *Index = NULL);
	NvMutex ProtectStartStop;		// protect locking for starting and stopping applications

public:
	std::vector<void *> m_vApplications;			// application list
	
	NvApplicationList();
	virtual ~NvApplicationList();

	// Removes all applications from the list
	void Clear(bool AlsoDefault = true);
	
	// Clears application registry caches
	void ClearRegistryCache();

	// Clears the application work directories
	void CleanWorkDir();
	
	// used to get an existing application
	// The function increments the in use counter of the corresponding application if found and if LockIt is set to true
	// If Name is NULL or blank, the function returns the default application
	NvApplication* GetApplication(const char *Name, bool LockIt = true);

	// Load the application list from registry
	void Load(bool NvdefOnly = false);

	// Saves the application list to the system registry
	bool Save();

	// Start applications
	void Start();

	// Stop applications
	void Stop();


	// Creates a new application with the given name if it doesn't exist
	// Fails if the application ever exists and is running
	bool CreateApplication(const char *Name, const char *Desc, int DefaultTimeOut = 300, const char *SecuApp = NULL, bool ScheduledTransac = true, bool AllowWebXml = true, bool FromLoad = false, bool AllowNotOwnedSession = true, bool CreateLicense = false, int AuthMode = NV_AUTHENTICATION_NIRVA, bool NirvaIsSsoUser = false, const char* ClientFilterString = NULL, const char* WebFilterString = NULL, int CommandClientFilterMode = NV_COMMAND_FILTER_NEVER, int CommandWebFilterMode = NV_COMMAND_FILTER_NEVER);


	// used to free an existing application
	// The function decrements the in use counter of the corresponding application
	void FreeApplication(const char *Name);
	
	// Remove the given application from the list
	bool RemoveApplication(const char *Name);

	// Enable an application (default)
	bool EnableApplication(const char *Name, bool FromLoad = false);

	// Start an application
	bool StartApplication(const char *Name);

	// Stop an application
	bool StopApplication(const char *Name, const char *SessionId);

	// Checks if a application exists in the list
	bool ApplicationExist(const char *Name);

	// Checks if a application is enabled or not
	bool IsApplicationEnable(const char *Name);

	// Get sessions ids connected to the application
	bool GetSessions(const char *Name, NvObjStringList *ConnectedSessions);

	// Get application info
	bool GetInfo(const char *Name, NvObjTable *ApplicationInfo);

	// Get application specific info
	bool GetInfoEx(const char *Name, NvObjIndStringList *ApplicationInfo);

	// Package application
	bool Package(const char *Name, const char *DescriptionFile, const char *PackageFile, std::string& Info);

	// Install application
	bool Install(const char *Name, const char *PackageFile, std::string& Info);

	// Start listeners
	bool StartListeners();

	// Stop listeners
	bool StopListeners();

	// CheckRequiredService
	bool IsRequiredService(std::string& ServiceName);
};



#endif //_APPLICATION_H
