// ***************************************************************************************
// File: base64.cpp
// Description: base64 encoding
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifdef MAC
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
#endif

#ifdef NV_SERVER
#include "nvs.h"
#else
#include "../nvc/nvcc.h"
#endif

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER


using namespace std;

bool FileBase64(FILE* FileSource, FILE* FileDest, bool Encode, bool WithNewLines)
{

	BIO* b64bio;
	BIO* rbio = NULL;
	BIO* wbio = NULL;


	// create two file BIOs
	if(!(rbio = BIO_new_fp(FileSource, BIO_NOCLOSE)))
		return false;
	if (!(wbio = BIO_new_fp(FileDest, BIO_NOCLOSE)))
	{
		BIO_free_all(rbio);
		return false;
	}

	// insert base64 filter

	if (!(b64bio=BIO_new(BIO_f_base64())))
	{
		BIO_free_all(wbio);
		BIO_free_all(rbio);
		return false;
	}
	if(Encode && (!WithNewLines))
		BIO_set_flags(b64bio, BIO_FLAGS_BASE64_NO_NL);

	if (Encode)
		wbio=BIO_push(b64bio,wbio);
	else
		rbio=BIO_push(b64bio,rbio);

	// pump stuff through filter

	int n;
	char *buffer;
	
	buffer = NVNEW char[4096];
	if(buffer == NULL)
	{
		BIO_free_all(wbio);
		BIO_free_all(rbio);
		return false;
	}

	while(1)
	{
		if((n=BIO_read(rbio,buffer,4096)) <= 0)
			break;
		if(BIO_write(wbio, buffer,n) != n)
		{
			delete buffer;
			BIO_free_all(wbio);
			BIO_free_all(rbio);
			return false;
		}
	}
	BIO_flush(wbio);
	delete buffer;

	BIO_free_all(wbio);
	BIO_free_all(rbio);
	return true;
}


bool FileBase64(const char* SourceFile, const char* DestFile, bool Encode, bool WithNewLines)
{
	FILE *FileSource;
	FILE *FileDest;
	
	if(SourceFile == NULL)
		return false;
	if(!strlen(SourceFile))
		return false;
	
	if(DestFile == NULL)
		return false;
	if(!strlen(DestFile))
		return false;

	FileSource = Nvfopen(SourceFile, "rb");
	if(FileSource == NULL)
		return false;

	FileDest = Nvfopen(DestFile, "wb");
	if(FileDest == NULL)
	{
		fclose(FileSource);
		return false;
	}

	bool RetVal = FileBase64(FileSource, FileDest, Encode, WithNewLines);
	fclose(FileDest);
	fclose(FileSource);
	
	return RetVal;
}


bool FileBase64Encode(FILE* FileSource, const char* DestFile, bool WithNewLines)
{
	FILE *FileDest;
	
	if(DestFile == NULL)
		return false;
	if(!strlen(DestFile))
		return false;

	FileDest = Nvfopen(DestFile, "wb");
	if(FileDest == NULL)
	{
		fclose(FileSource);
		return false;
	}

	bool RetVal = FileBase64(FileSource, FileDest, true, WithNewLines);
	fclose(FileDest);
	
	return RetVal;
}

bool FileBase64Decode(FILE* FileSource, const char* DestFile)
{
	FILE *FileDest;
	
	if(DestFile == NULL)
		return false;
	if(!strlen(DestFile))
		return false;

	FileDest = Nvfopen(DestFile, "wb");
	if(FileDest == NULL)
	{
		fclose(FileSource);
		return false;
	}

	bool RetVal = FileBase64(FileSource, FileDest, false, false);
	fclose(FileDest);
	
	return RetVal;
}


bool FileBase64Encode(const char* SourceFile, FILE* FileDest, bool WithNewLines)
{
	FILE *FileSource;
	
	if(SourceFile == NULL)
		return false;
	if(!strlen(SourceFile))
		return false;
	

	FileSource = Nvfopen(SourceFile, "rb");
	if(FileSource == NULL)
		return false;

	bool RetVal = FileBase64(FileSource, FileDest, true, WithNewLines);
	fclose(FileSource);
	
	return RetVal;
}

bool FileBase64Decode(const char* SourceFile, FILE* FileDest)
{
	FILE *FileSource;
	
	if(SourceFile == NULL)
		return false;
	if(!strlen(SourceFile))
		return false;
	

	FileSource = Nvfopen(SourceFile, "rb");
	if(FileSource == NULL)
		return false;

	bool RetVal = FileBase64(FileSource, FileDest, false, false);
	fclose(FileSource);
	
	return RetVal;
}


bool FileBase64Encode(const char* SourceFile, const char* DestFile, bool WithNewLines)
{
	return FileBase64(SourceFile, DestFile, true, WithNewLines);
}

bool FileBase64Decode(const char* SourceFile, const char* DestFile)
{
	return FileBase64(SourceFile, DestFile, false, false);
}

bool MemBase64(const char* SourceData, char** DestinationData, int* DataSize, bool Encode, bool WithNewLines)
{
	if(SourceData == NULL)
		return false;
	if(*DataSize < 0)
		return false;

	int SourceDataSize = *DataSize;
	int NumBlocks = SourceDataSize / 512;
	int LastBlock = SourceDataSize % 512;

	BIO* b64bio;
	BIO* rbio = NULL;
	BIO* wbio = NULL;
	char* pd;

	*DataSize = 0;
	if(DestinationData)
	{
		if(*DestinationData != NULL)
			free(*DestinationData);
		*DestinationData = NULL;
	}

	// create two memory buffer BIOs
	if(!(rbio = BIO_new_mem_buf((void *)SourceData, SourceDataSize)))
		return false;
	if (!(wbio = BIO_new(BIO_s_mem())))
	{
		BIO_free_all(rbio);
		return false;
	}

	// insert base64 filter

	if (!(b64bio=BIO_new(BIO_f_base64())))
	{
		BIO_free_all(wbio);
		BIO_free_all(rbio);
		return false;
	}

	if(!Encode)
	{
		// verify if the input encode base64 data contains line feed (normaly max line is 76 chars so we can check for 99 bytes max)
		int Check_size = 99;
		if(strlen(SourceData) < 99)
			Check_size = (int)(strlen(SourceData));
		char BufTmp[100];
		memcpy(BufTmp, SourceData, Check_size);
		BufTmp[Check_size] = '\0';
		string StartString = BufTmp;
		if(StartString.find('\n') == string::npos)
			BIO_set_flags(b64bio, BIO_FLAGS_BASE64_NO_NL);
	}
	if(Encode && (!WithNewLines))
		BIO_set_flags(b64bio, BIO_FLAGS_BASE64_NO_NL);


	if (Encode)
		wbio=BIO_push(b64bio,wbio);
	else
		rbio=BIO_push(b64bio,rbio);

	// pump stuff through filter

	int n;
	char *buffer;
	
	buffer = NVNEW char[4096];
	if(buffer == NULL)
	{
		BIO_free_all(wbio);
		BIO_free_all(rbio);
		return false;
	}
	while(1)
	{
		if((n=BIO_read(rbio,buffer,4096)) <= 0)
			break;
		if(BIO_write(wbio, buffer,n) != n)
		{
			delete buffer;
			BIO_free_all(wbio);
			BIO_free_all(rbio);
			return false;
		}
	}
	BIO_flush(wbio);
	delete buffer;

	// free BIOs and return base64 encoded block
	int i=(int)BIO_ctrl(wbio,BIO_CTRL_INFO,0,(char *)&pd);
	if(i>0)
	{
		*DestinationData = (char*)malloc(i+1);	// We add one character for strings
		memcpy(*DestinationData, (const void*)pd, i);
		*DataSize = i;
		(*DestinationData)[i] = '\0';
	}
	else
	{
		BIO_free_all(wbio);
		BIO_free_all(rbio);
		return false;
	}

	BIO_free_all(wbio);
	BIO_free_all(rbio);
	return true;
}

bool MemBase64Encode(const char* SourceData, char** DestinationData, int* DataSize, bool WithNewLines)
{
	return MemBase64(SourceData, DestinationData, DataSize, true, WithNewLines);
}

bool MemBase64Decode(const char* SourceData, char** DestinationData, int* DataSize)
{
	return MemBase64(SourceData, DestinationData, DataSize, false, false);
}

bool FileMemBase64(FILE* FileSource, char** DestinationData, int* DataSize, bool Encode, bool WithNewLines)
{
	if(DestinationData == NULL)
		return false;
	if(*DataSize < 0)
		return false;

	*DataSize = 0;
	if(DestinationData)
	{
		if(*DestinationData != NULL)
			free(*DestinationData);
		*DestinationData = NULL;
	}

	BIO* b64bio;
	BIO* rbio = NULL;
	BIO* wbio = NULL;
	char* pd;


	// create 1 file BIO
	if(!(rbio = BIO_new_fp(FileSource, BIO_NOCLOSE)))
		return false;

	// create 1 memory buffer BIO
	if (!(wbio = BIO_new(BIO_s_mem())))
	{
		BIO_free_all(rbio);
		return false;
	}

	// insert base64 filter

	if (!(b64bio=BIO_new(BIO_f_base64())))
	{
		BIO_free_all(wbio);
		BIO_free_all(rbio);
		return false;
	}
	if(Encode && (!WithNewLines))
		BIO_set_flags(b64bio, BIO_FLAGS_BASE64_NO_NL);

	if (Encode)
		wbio=BIO_push(b64bio,wbio);
	else
		rbio=BIO_push(b64bio,rbio);

	// pump stuff through filter

	int n;
	char *buffer;
	
	buffer = NVNEW char[4096];
	if(buffer == NULL)
	{
		BIO_free_all(wbio);
		BIO_free_all(rbio);
		return false;
	}
	while(1)
	{
		if((n=BIO_read(rbio,buffer,4096)) <= 0)
			break;
		if(BIO_write(wbio, buffer,n) != n)
		{
			delete buffer;
			BIO_free_all(wbio);
			BIO_free_all(rbio);
			return false;
		}
	}
	BIO_flush(wbio);
	delete buffer;

	// free BIOs and return base64 encoded block
	int i=(int)BIO_ctrl(wbio,BIO_CTRL_INFO,0,(char *)&pd);
	if(i>0)
	{
		*DestinationData = (char*)malloc(i+1);	// We add one character for strings
		memcpy(*DestinationData, (const void*)pd, i);
		*DataSize = i;
		(*DestinationData)[i] = '\0';
	}
	else
	{
		BIO_free_all(wbio);
		BIO_free_all(rbio);
		return false;
	}

	BIO_free_all(wbio);
	BIO_free_all(rbio);
	return true;
}

bool FileMemBase64(const char* SourceFile, char** DestinationData, int* DataSize, bool Encode, bool WithNewLines)
{
	FILE *FileSource;
	
	if(SourceFile == NULL)
		return false;
	if(!strlen(SourceFile))
		return false;

	FileSource = Nvfopen(SourceFile, "rb");
	if(FileSource == NULL)
		return false;

	bool RetVal = FileMemBase64(FileSource, DestinationData, DataSize, Encode, WithNewLines);
	fclose(FileSource);
	return RetVal;
}

bool FileMemBase64Encode(FILE* FileSource, char** DestinationData, int* DataSize, bool WithNewLines)
{
	return FileMemBase64(FileSource, DestinationData, DataSize, true, WithNewLines);
}

bool FileMemBase64Decode(FILE* FileSource, char** DestinationData, int* DataSize)
{
	return FileMemBase64(FileSource, DestinationData, DataSize, false, false);
}

bool FileMemBase64Encode(const char* SourceFile, char** DestinationData, int* DataSize, bool WithNewLines)
{
	return FileMemBase64(SourceFile, DestinationData, DataSize, true, WithNewLines);
}

bool FileMemBase64Decode(const char* SourceFile, char** DestinationData, int* DataSize)
{
	return FileMemBase64(SourceFile, DestinationData, DataSize, false, false);
}


bool MemFileBase64(const char* SourceData, int DataSize, FILE* FileDest, bool Encode, bool WithNewLines)
{
	if(SourceData == NULL)
		return false;
	if(DataSize < 0)
		return false;

	int SourceDataSize = DataSize;
	int NumBlocks = SourceDataSize / 512;
	int LastBlock = SourceDataSize % 512;

	BIO* b64bio;
	BIO* rbio = NULL;
	BIO* wbio = NULL;

	// create 1 memory buffer BIO
	if(!(rbio = BIO_new_mem_buf((void *)SourceData, SourceDataSize)))
		return false;
	// create 1 file BIO
	if (!(wbio = BIO_new_fp(FileDest, BIO_NOCLOSE)))
	{
		BIO_free_all(rbio);
		return false;
	}

	// insert base64 filter

	if (!(b64bio=BIO_new(BIO_f_base64())))
	{
		BIO_free_all(wbio);
		BIO_free_all(rbio);
		return false;
	}
	
	if(!Encode)
	{
		// verify if the input encode base64 data contains line feed (normaly max line is 76 chars so we can check for 99 bytes max)
		int Check_size = 99;
		if(strlen(SourceData) < 99)
			Check_size = (int)(strlen(SourceData));
		char BufTmp[100];
		memcpy(BufTmp, SourceData, Check_size);
		BufTmp[Check_size] = '\0';
		string StartString = BufTmp;
		if(StartString.find('\n') == string::npos)
			BIO_set_flags(b64bio, BIO_FLAGS_BASE64_NO_NL);
	}
	
	if(Encode && (!WithNewLines))
		BIO_set_flags(b64bio, BIO_FLAGS_BASE64_NO_NL);

	if (Encode)
		wbio=BIO_push(b64bio,wbio);
	else
		rbio=BIO_push(b64bio,rbio);

	// pump stuff through filter

	int n;
	char *buffer;
	
	buffer = NVNEW char[4096];
	if(buffer == NULL)
	{
		BIO_free_all(wbio);
		BIO_free_all(rbio);
		return false;
	}

	while(1)
	{
		if((n=BIO_read(rbio,buffer,4096)) <= 0)
			break;
		if(BIO_write(wbio, buffer,n) != n)
		{
			delete buffer;
			BIO_free_all(wbio);
			BIO_free_all(rbio);
			return false;
		}
	}
	BIO_flush(wbio);
	delete buffer;


	BIO_free_all(wbio);
	BIO_free_all(rbio);
	return true;
}


bool MemFileBase64(const char* SourceData, int DataSize, const char* DestFile, bool Encode, bool WithNewLines)
{
	FILE *FileDest;
	
	if(DestFile == NULL)
		return false;
	if(!strlen(DestFile))
		return false;

	FileDest = Nvfopen(DestFile, "wb");
	if(FileDest == NULL)
		return false;

	bool RetVal = MemFileBase64(SourceData, DataSize, FileDest, Encode, WithNewLines);
	fclose(FileDest);
	return RetVal;
}

bool MemFileBase64Encode(const char* SourceData, int DataSize, FILE* FileDest, bool WithNewLines)
{
	return MemFileBase64(SourceData, DataSize, FileDest, true, WithNewLines);
}

bool MemFileBase64Decode(const char* SourceData, int DataSize, FILE* FileDest)
{
	return MemFileBase64(SourceData, DataSize, FileDest, false, false);
}

bool MemFileBase64Encode(const char* SourceData, int DataSize, const char* DestFile, bool WithNewLines)
{
	return MemFileBase64(SourceData, DataSize, DestFile, true, WithNewLines);
}

bool MemFileBase64Decode(const char* SourceData, int DataSize, const char* DestFile)
{
	return MemFileBase64(SourceData, DataSize, DestFile, false, false);
}
